library angular2.src.change_detection.change_detection;

import "jit_proto_change_detector.dart" show JitProtoChangeDetector;
import "pregen_proto_change_detector.dart" show PregenProtoChangeDetector;
import "proto_change_detector.dart" show DynamicProtoChangeDetector;
import "pipes/pipe.dart" show PipeFactory, Pipe;
import "pipes/pipe_registry.dart" show PipeRegistry;
import "pipes/iterable_changes.dart" show IterableChangesFactory;
import "pipes/keyvalue_changes.dart" show KeyValueChangesFactory;
import "pipes/observable_pipe.dart" show ObservablePipeFactory;
import "pipes/promise_pipe.dart" show PromisePipeFactory;
import "pipes/uppercase_pipe.dart" show UpperCaseFactory;
import "pipes/lowercase_pipe.dart" show LowerCaseFactory;
import "pipes/json_pipe.dart" show JsonPipe;
import "pipes/null_pipe.dart" show NullPipeFactory;
import "interfaces.dart"
    show ChangeDetection, ProtoChangeDetector, ChangeDetectorDefinition;
import "package:angular2/di.dart"
    show Inject, Injectable, OpaqueToken, Optional;
import "package:angular2/src/facade/collection.dart"
    show List, Map, StringMapWrapper;
import "package:angular2/src/facade/lang.dart" show isPresent, BaseException;

/**
 * Structural diffing for `Object`s and `Map`s.
 *
 * @exportedAs angular2/pipes
 */
List<PipeFactory> keyValDiff = [
  new KeyValueChangesFactory(),
  new NullPipeFactory()
];
/**
 * Structural diffing for `Iterable` types such as `Array`s.
 *
 * @exportedAs angular2/pipes
 */
List<PipeFactory> iterableDiff = [
  new IterableChangesFactory(),
  new NullPipeFactory()
];
/**
 * Async binding to such types as Observable.
 *
 * @exportedAs angular2/pipes
 */
List<PipeFactory> async = [
  new ObservablePipeFactory(),
  new PromisePipeFactory(),
  new NullPipeFactory()
];
/**
 * Uppercase text transform.
 *
 * @exportedAs angular2/pipes
 */
List<PipeFactory> uppercase = [new UpperCaseFactory(), new NullPipeFactory()];
/**
 * Lowercase text transform.
 *
 * @exportedAs angular2/pipes
 */
List<PipeFactory> lowercase = [new LowerCaseFactory(), new NullPipeFactory()];
/**
 * Json stringify transform.
 *
 * @exportedAs angular2/pipes
 */
List<dynamic /* PipeFactory | Pipe */ > json = [
  new JsonPipe(),
  new NullPipeFactory()
];
var defaultPipes = {
  "iterableDiff": iterableDiff,
  "keyValDiff": keyValDiff,
  "async": async,
  "uppercase": uppercase,
  "lowercase": lowercase,
  "json": json
};
/**
 * Map from {@link ChangeDetectorDefinition#id} to a factory method which takes a
 * {@link PipeRegistry} and a {@link ChangeDetectorDefinition} and generates a
 * {@link ProtoChangeDetector} associated with the definition.
 */

// TODO(kegluneq): Use PregenProtoChangeDetectorFactory rather than Function once possible in

// dart2js. See https://github.com/dart-lang/sdk/issues/23630 for details.
Map<String, Function> preGeneratedProtoDetectors = {};
const PROTO_CHANGE_DETECTOR_KEY = const OpaqueToken("ProtoChangeDetectors");
/**
 * Implements change detection using a map of pregenerated proto detectors.
 *
 * @exportedAs angular2/change_detection
 */
@Injectable()
class PreGeneratedChangeDetection extends ChangeDetection {
  PipeRegistry registry;
  ChangeDetection _dynamicChangeDetection;
  Map<String, Function> _protoChangeDetectorFactories;
  PreGeneratedChangeDetection(this.registry, [@Inject(
      PROTO_CHANGE_DETECTOR_KEY) @Optional() Map<String, Function> protoChangeDetectorsForTest])
      : super() {
    /* super call moved to initializer */;
    this._dynamicChangeDetection = new DynamicChangeDetection(registry);
    this._protoChangeDetectorFactories = isPresent(protoChangeDetectorsForTest)
        ? protoChangeDetectorsForTest
        : preGeneratedProtoDetectors;
  }
  static bool isSupported() {
    return PregenProtoChangeDetector.isSupported();
  }
  ProtoChangeDetector createProtoChangeDetector(
      ChangeDetectorDefinition definition) {
    var id = definition.id;
    if (StringMapWrapper.contains(this._protoChangeDetectorFactories, id)) {
      return StringMapWrapper.get(this._protoChangeDetectorFactories, id)(
          this.registry, definition);
    }
    return this._dynamicChangeDetection.createProtoChangeDetector(definition);
  }
}
/**
 * Implements change detection that does not require `eval()`.
 *
 * This is slower than {@link JitChangeDetection}.
 *
 * @exportedAs angular2/change_detection
 */
@Injectable()
class DynamicChangeDetection extends ChangeDetection {
  PipeRegistry registry;
  DynamicChangeDetection(this.registry) : super() {
    /* super call moved to initializer */;
  }
  ProtoChangeDetector createProtoChangeDetector(
      ChangeDetectorDefinition definition) {
    return new DynamicProtoChangeDetector(this.registry, definition);
  }
}
/**
 * Implements faster change detection by generating source code.
 *
 * This requires `eval()`. For change detection that does not require `eval()`, see
 * {@link DynamicChangeDetection} and {@link PreGeneratedChangeDetection}.
 *
 * @exportedAs angular2/change_detection
 */
@Injectable()
class JitChangeDetection extends ChangeDetection {
  PipeRegistry registry;
  JitChangeDetection(this.registry) : super() {
    /* super call moved to initializer */;
  }
  static bool isSupported() {
    return JitProtoChangeDetector.isSupported();
  }
  ProtoChangeDetector createProtoChangeDetector(
      ChangeDetectorDefinition definition) {
    return new JitProtoChangeDetector(this.registry, definition);
  }
}
PipeRegistry defaultPipeRegistry = new PipeRegistry(defaultPipes);
