library angular2.src.router.router;

import "package:angular2/src/facade/async.dart"
    show Future, PromiseWrapper, EventEmitter, ObservableWrapper;
import "package:angular2/src/facade/collection.dart"
    show Map, MapWrapper, List, ListWrapper;
import "package:angular2/src/facade/lang.dart" show isBlank, isPresent, Type;
import "route_registry.dart" show RouteRegistry;
import "pipeline.dart" show Pipeline;
import "instruction.dart" show Instruction;
import "router_outlet.dart" show RouterOutlet;
import "location.dart" show Location;

/**
 * # Router
 * The router is responsible for mapping URLs to components.
 *
 * You can see the state of the router by inspecting the read-only field `router.navigating`.
 * This may be useful for showing a spinner, for instance.
 *
 * ## Concepts
 * Routers and component instances have a 1:1 correspondence.
 *
 * The router holds reference to a number of "outlets." An outlet is a placeholder that the
 * router dynamically fills in depending on the current URL.
 *
 * When the router navigates from a URL, it must first recognizes it and serialize it into an
 * `Instruction`.
 * The router uses the `RouteRegistry` to get an `Instruction`.
 *
 * @exportedAs angular2/router
 */
class Router {
  RouteRegistry _registry;
  Pipeline _pipeline;
  Router parent;
  dynamic hostComponent;
  bool navigating;
  String lastNavigationAttempt;
  String previousUrl;
  Instruction _currentInstruction;
  Future<dynamic> _currentNavigation;
  RouterOutlet _outlet;
  EventEmitter _subject;
  // todo(jeffbcross): rename _registry to registry since it is accessed from subclasses

  // todo(jeffbcross): rename _pipeline to pipeline since it is accessed from subclasses
  Router(this._registry, this._pipeline, this.parent, this.hostComponent) {
    this.navigating = false;
    this.previousUrl = null;
    this._outlet = null;
    this._subject = new EventEmitter();
    this._currentInstruction = null;
    this._currentNavigation = PromiseWrapper.resolve(true);
  }
  /**
   * Constructs a child router. You probably don't need to use this unless you're writing a reusable
   * component.
   */
  Router childRouter(dynamic hostComponent) {
    return new ChildRouter(this, hostComponent);
  }
  /**
   * Register an object to notify of route changes. You probably don't need to use this unless
   * you're writing a reusable component.
   */
  Future<bool> registerOutlet(RouterOutlet outlet) {
    // TODO: sibling routes
    this._outlet = outlet;
    if (isPresent(this._currentInstruction)) {
      return outlet.activate(this._currentInstruction);
    }
    return PromiseWrapper.resolve(true);
  }
  /**
   * Dynamically update the routing configuration and trigger a navigation.
   *
   * # Usage
   *
   * ```
   * router.config({ 'path': '/', 'component': IndexCmp});
   * ```
   *
   * Or:
   *
   * ```
   * router.config([
   *   { 'path': '/', 'component': IndexComp },
   *   { 'path': '/user/:id', 'component': UserComp },
   * ]);
   * ```
   */
  Future<dynamic> config(dynamic config) {
    if (config is List) {
      ((config as List<dynamic>)).forEach((configObject) {
        this._registry.config(this.hostComponent, configObject);
      });
    } else {
      this._registry.config(this.hostComponent, config);
    }
    return this.renavigate();
  }
  /**
   * Navigate to a URL. Returns a promise that resolves when navigation is complete.
   *
   * If the given URL begins with a `/`, router will navigate absolutely.
   * If the given URL does not begin with `/`, the router will navigate relative to this component.
   */
  Future<dynamic> navigate(String url) {
    if (this.navigating) {
      return this._currentNavigation;
    }
    this.lastNavigationAttempt = url;
    return this._currentNavigation = this
        .recognize(url)
        .then((matchedInstruction) {
      if (isBlank(matchedInstruction)) {
        return PromiseWrapper.resolve(false);
      }
      if (isPresent(this._currentInstruction)) {
        matchedInstruction.reuseComponentsFrom(this._currentInstruction);
      }
      this._startNavigating();
      var result = this.commit(matchedInstruction).then((_) {
        this._finishNavigating();
        ObservableWrapper.callNext(
            this._subject, matchedInstruction.accumulatedUrl);
      });
      return PromiseWrapper.catchError(result, (err) {
        this._finishNavigating();
        throw err;
      });
    });
  }
  void _startNavigating() {
    this.navigating = true;
  }
  void _finishNavigating() {
    this.navigating = false;
  }
  /**
   * Subscribe to URL updates from the router
   */
  void subscribe(onNext) {
    ObservableWrapper.subscribe(this._subject, onNext);
  }
  /**
   * Updates this router and all descendant routers according to the given instruction
   */
  Future<dynamic> commit(Instruction instruction) {
    this._currentInstruction = instruction;
    if (isPresent(this._outlet)) {
      return this._outlet.activate(instruction);
    }
    return PromiseWrapper.resolve(true);
  }
  /**
   * Removes the contents of this router's outlet and all descendant outlets
   */
  Future<dynamic> deactivate() {
    if (isPresent(this._outlet)) {
      return this._outlet.deactivate();
    }
    return PromiseWrapper.resolve(true);
  }
  /**
   * Given a URL, returns an instruction representing the component graph
   */
  Future<Instruction> recognize(String url) {
    return this._registry.recognize(url, this.hostComponent);
  }
  /**
   * Navigates to either the last URL successfully navigated to, or the last URL requested if the
   * router has yet to successfully navigate.
   */
  Future<dynamic> renavigate() {
    var destination = isBlank(this.previousUrl)
        ? this.lastNavigationAttempt
        : this.previousUrl;
    if (isBlank(destination)) {
      return this._currentNavigation;
    }
    return this.navigate(destination);
  }
  /**
   * Generate a URL from a component name and optional map of parameters. The URL is relative to the
   * app's base href.
   */
  String generate(String name, Map<String, String> params) {
    return this._registry.generate(name, params, this.hostComponent);
  }
}
class RootRouter extends Router {
  Location _location;
  RootRouter(RouteRegistry registry, Pipeline pipeline, Location location,
      Type hostComponent)
      : super(registry, pipeline, null, hostComponent) {
    /* super call moved to initializer */;
    this._location = location;
    this._location.subscribe((change) => this.navigate(change["url"]));
    this._registry.configFromComponent(hostComponent);
    this.navigate(location.path());
  }
  Future<dynamic> commit(instruction) {
    return super.commit(instruction).then((_) {
      this._location.go(instruction.accumulatedUrl);
    });
  }
}
class ChildRouter extends Router {
  ChildRouter(Router parent, hostComponent)
      : super(parent._registry, parent._pipeline, parent, hostComponent) {
    /* super call moved to initializer */;
    this.parent = parent;
  }
}
